# Copyright (c) 2022 Exograd SAS.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import datetime
from typing import Any, Dict, List, Optional, TypeVar, Type


import dateutil.parser

FieldType = TypeVar("FieldType", str, datetime.datetime, int, bool, dict, list)


class InvalidAPIObjectError(Exception):
    """An error signaled when an API object contains invalid data."""

    def __init__(self, object_name: str, value: Any, reason: str) -> None:
        super().__init__(f"invalid {object_name}: {reason}")

        self.object_name = object_name
        self.value = value
        self.reason = reason


class APIObject:
    """An object exposed by the Eventline API."""

    def __init__(self, object_name: str):
        self._object_name = object_name

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        string = f"<eventline.{self._object_name}"
        if hasattr(self, "id_"):
            id_ = getattr(self, "id_")
            if id_ is not None:
                string += f" {id_}"
        string += ">"
        return string


class ReadableAPIObject(APIObject):
    """An API object which can be read from a JSON object."""

    def _read(self, data: Dict[str, Any]) -> None:
        pass

    def _get_optional_field(
        self,
        data: Dict[str, Any],
        key: str,
        class_type: Type[FieldType],
        class_name: str,
    ) -> Optional[FieldType]:
        if not isinstance(data, dict):
            raise InvalidAPIObjectError(
                "response", data, "response data are not an object"
            )
        if key not in data:
            return None
        value = data.get(key, None)
        if value is not None and not isinstance(value, class_type):
            article = "a"
            if class_name[0] in ("a", "e", "i", "o", "u"):
                article = "an"
            raise InvalidAPIObjectError(
                self._object_name,
                value,
                f"field '{key}' is not {article} {class_name}",
            )
        return value

    def _get_field(
        self,
        data: Dict[str, Any],
        key: str,
        class_type: Type[FieldType],
        class_name: str,
    ) -> FieldType:
        value = self._get_optional_field(data, key, class_type, class_name)
        if value is None:
            raise InvalidAPIObjectError(
                self._object_name, data, f"missing field '{key}'"
            )
        return value

    def _read_optional_string(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> Optional[str]:
        return self._get_optional_field(data, key, str, "string")

    def _read_string(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> str:
        return self._get_field(data, key, str, "string")

    def _read_optional_datetime(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> Optional[datetime.datetime]:
        string = self._get_field(data, key, str, "string")
        value = None
        if string is not None:
            try:
                value = dateutil.parser.isoparse(string)
            except Exception as ex:
                raise InvalidAPIObjectError(
                    self._object_name,
                    string,
                    f"field '{key}' is not a valid datetime",
                ) from ex
        return value

    def _read_datetime(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> datetime.datetime:
        value = self._read_optional_datetime(data, key)
        if value is None:
            raise InvalidAPIObjectError(
                self._object_name, data, f"missing field '{key}'"
            )
        return value

    def _read_optional_integer(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> Optional[int]:
        return self._get_optional_field(data, key, int, "integer")

    def _read_integer(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> int:
        return self._get_field(data, key, int, "integer")

    def _read_optional_boolean(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> Optional[bool]:
        value = self._get_optional_field(data, key, bool, "boolean")
        if value is None:
            value = False
        return value

    def _read_boolean(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> bool:
        return self._get_field(data, key, bool, "boolean")

    def _read_optional_object(
        self,
        data: Dict[str, Any],
        key: str,
        class_type: Any,
    ) -> Optional[object]:
        obj = self._get_optional_field(data, key, dict, "object")
        value = None
        if obj is not None:
            value = class_type()
            value._read(obj)
        return value

    def _read_object(
        self,
        data: Dict[str, Any],
        key: str,
        class_type: Any,
    ) -> object:
        value = self._read_optional_object(data, key, class_type)
        if value is None:
            raise InvalidAPIObjectError(
                self._object_name, data, f"missing field '{key}'"
            )
        return value

    def _read_optional_object_array(
        self,
        data: Dict[str, Any],
        key: str,
        element_class_type: Any,
    ) -> Optional[List[object]]:
        array = self._get_field(data, key, list, "array")
        value = None
        if array is not None:
            value = []
            for i, element in enumerate(array):
                if not isinstance(element, dict):
                    raise InvalidAPIObjectError(
                        self._object_name,
                        element,
                        f"element at index {i} of field '{key}' is not an "
                        "object",
                    )
                element_value = element_class_type()
                element_value._read(element)
                value.append(element_value)
        return value

    def _read_object_array(
        self,
        data: Dict[str, Any],
        key: str,
        element_class_type: Any,
    ) -> List[object]:
        value = self._read_optional_object_array(data, key, element_class_type)
        if value is None:
            raise InvalidAPIObjectError(
                self._object_name, data, f"missing field '{key}'"
            )
        return value

    def _read_optional_string_array(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> Optional[List[str]]:
        array = self._get_field(data, key, list, "array")
        value = None
        if array is not None:
            value = []
            for i, element in enumerate(array):
                if not isinstance(element, str):
                    raise InvalidAPIObjectError(
                        self._object_name,
                        element,
                        f"element at index {i} of field '{key}' is not a "
                        "string",
                    )
                value.append(element)
        return value

    def _read_string_array(
        self,
        data: Dict[str, Any],
        key: str,
    ) -> List[str]:
        value = self._read_optional_string_array(data, key)
        if value is None:
            raise InvalidAPIObjectError(
                self._object_name, data, f"missing field '{key}'"
            )
        return value


class SerializableAPIObject(APIObject):
    """An API object which can be serialized to a JSON object."""

    def _serialize(self) -> Dict[str, Any]:
        return {}



# Copyright (c) 2022 Exograd SAS.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import hashlib
import json
import json.decoder
import logging
import time
from typing import Any, Dict, Optional
import urllib.parse

import OpenSSL.crypto
import urllib3

import eventline.environment
import eventline.pagination
from eventline.project import Project

log = logging.getLogger(__name__)


class ClientError(Exception):
    """An error encountered by the client."""


class APIError(ClientError):
    """An error signaled when a request failed due to an API error."""

    def __init__(
        self,
        method: str,
        uri: str,
        status: int,
        /,
        error_message: Optional[str],
        error_code: Optional[str],
    ) -> None:
        message = f"{method} {uri}: request failed with status {status}"
        if error_message is not None:
            message += f": {error_message}"

        super().__init__(message)

        self.method = method
        self.uri = uri
        self.status = status
        self.error_code = error_code
        self.error_message = error_message


class Response:
    """A representation of a successful response returned by Eventline."""

    def __init__(self, response: urllib3.HTTPResponse) -> None:
        self.status = response.status
        self.header = response.headers
        self.raw_body = response.data
        self.body = self._decode_body()

    def _decode_body(self) -> Any:
        if len(self.raw_body) == 0:
            return None
        content_type = self.header.get("Content-Type")
        if content_type == "application/json":
            return self._decode_json_body()
        if content_type == "text/plain":
            return self.raw_body.decode("UTF-8", "strict")
        raise ClientError(f"unhandled content type '{content_type}'")

    def _decode_json_body(self) -> Any:
        try:
            return json.loads(self.raw_body)
        except json.decoder.JSONDecodeError as ex:
            raise ClientError(f"cannot decode response body: {ex}") from ex


class Client:
    """The low level HTTP client for the Eventline API."""

    default_endpoint = "https://api.eventline.net/v0"

    def __init__(
        self,
        /,
        endpoint: str = default_endpoint,
        timeout: float = 10.0,
        api_key: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        self.endpoint = urllib.parse.urlparse(endpoint)
        scheme = self.endpoint.scheme.lower()
        if not scheme in ("http", "https"):
            raise ClientError(
                f"invalid endpoint uri scheme '{self.endpoint.scheme}'"
            )

        self.api_key = api_key
        if self.api_key is None:
            self.api_key = eventline.environment.api_key()

        host = self.endpoint.hostname
        port = self.endpoint.port
        if port is None:
            if scheme == "http":
                port = 80
            else:
                port = 443

        if scheme == "http":
            self.pool = urllib3.HTTPConnectionPool(
                host,
                port=port,
                timeout=timeout,
                retries=0,
            )
        else:
            self.pool = HTTPSConnectionPool(
                host,
                port=port,
                ca_certs=eventline.ca_bundle_path,
                timeout=timeout,
                retries=0,
            )

        self.project_id = None  # type: Optional[str]
        if project_id is not None and project_name is not None:
            raise ClientError("cannot set both project_id and project_name")
        if project_id is not None:
            self.project_id = project_id
        elif project_name is not None:
            response = self.send_request(
                "GET", f"/projects/name/{urllib.parse.quote(project_name)}"
            )
            project = Project()
            project._read(response.body)
            self.project_id = project.id_
        else:
            self.project_id = eventline.environment.project_id()

    def send_request(
        self,
        method: str,
        path: str,
        /,
        query_parameters: Optional[Dict[str, str]] = None,
        body: Optional[Any] = None,
        raw_body: bool = False,
        cursor: Optional[eventline.pagination.Cursor] = None,
    ) -> Response:
        """Send a HTTP request and return the response.

        Raise an APIError exception if the status code of the response
        indicates an error.
        """
        uri = self.build_uri(path, query_parameters, cursor)
        headers = self._build_headers(body)
        try:
            body_data = None
            if body is not None:
                if raw_body is True:
                    body_data = body
                else:
                    body_data = json.dumps(body)
            start = time.monotonic()
            response = self.pool.urlopen(
                method,
                uri,
                headers=headers,
                body=body_data,
            )
            end = time.monotonic()
        except Exception as ex:
            raise ClientError(ex) from ex
        status = response.status
        time_string = format_request_time(end - start)
        log.debug(f"{method} {uri} {status} {time_string}")
        self._check_response(method, uri, response)
        return Response(response)

    def build_uri(
        self,
        path: str,
        query_parameters: Optional[Dict[str, str]] = None,
        cursor: Optional[eventline.pagination.Cursor] = None,
    ) -> str:
        """Construct the URI for an Eventline API route."""
        scheme, address, base_path, _, _, _ = self.endpoint
        full_path = base_path
        if full_path.endswith("/"):
            full_path = full_path[:-1]  # String.removesuffix is 3.9+
        full_path += path
        query = ""
        all_query_parameters = self._build_query_parameters(
            query_parameters, cursor
        )
        if len(all_query_parameters) > 0:
            query = urllib.parse.urlencode(all_query_parameters)
        fragment = ""
        components = (scheme, address, full_path, "", query, fragment)
        return urllib.parse.urlunparse(components)

    def _build_query_parameters(
        self,
        query_parameters: Optional[Dict[str, str]] = None,
        cursor: Optional[eventline.pagination.Cursor] = None,
    ) -> Dict[str, str]:
        all_query_parameters = {}
        if query_parameters is not None:
            all_query_parameters.update(query_parameters)
        if cursor is not None:
            if cursor.before is not None:
                all_query_parameters["before"] = cursor.before
            if cursor.after is not None:
                all_query_parameters["after"] = cursor.after
            if cursor.size is not None:
                all_query_parameters["size"] = str(cursor.size)
            if cursor.sort is not None:
                all_query_parameters["sort"] = cursor.sort
            if cursor.order is not None:
                all_query_parameters["order"] = cursor.order
        return all_query_parameters

    def _build_headers(self, /, body: Optional[Any]) -> Dict[str, str]:
        """Build the set of header fields for a request."""

        headers = {}

        if self.api_key is not None:
            headers["Authorization"] = "Bearer " + self.api_key

        if self.project_id is not None:
            headers["X-Eventline-Project-Id"] = self.project_id

        if body is not None:
            headers["Content-Type"] = "application/json"

        return headers

    def _check_response(
        self, method: str, uri: str, response: urllib3.HTTPResponse
    ) -> None:
        """Check if a response indicates success or failure, and signal an
        APIError exception if it is a failure."""

        status = response.status
        if 200 <= status < 300:
            return

        message = response.reason
        code = None

        try:
            error = json.loads(response.data)
            message = error["error"]
            code = error["code"]
        except json.decoder.JSONDecodeError:
            pass
        raise APIError(
            method,
            uri,
            response.status,
            error_message=message,
            error_code=code,
        )


def format_request_time(seconds: float) -> str:
    """Format and return the time used to send a request and obtain the
    response."""
    if seconds < 0.001:
        return f"{seconds*1_000_000:.0f}μs"
    if seconds < 1.0:
        return f"{seconds*1_000:.0f}ms"
    return f"{seconds:0.1f}s"


class HTTPSConnectionPool(urllib3.HTTPSConnectionPool):
    """An urllib3 connection pool which performs public key pinning."""

    def _validate_conn(self, conn: urllib3.connection.HTTPConnection) -> None:
        # Ignore mypy error: "_validate_conn" undefined in superclass.
        # This is obviously wrong.
        super()._validate_conn(conn)  # type: ignore
        if not conn.is_verified:
            return

        cert_data = conn.sock.getpeercert(binary_form=True)
        cert = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_ASN1, cert_data
        )

        key = cert.get_pubkey()
        key_data = OpenSSL.crypto.dump_publickey(
            OpenSSL.crypto.FILETYPE_ASN1, key
        )

        pin = hashlib.sha256(key_data).hexdigest()

        if pin not in eventline.public_key_pins:
            raise ClientError(
                f"invalid server certificate: unknown public key (pin {pin})"
            )



# Copyright (c) 2022 Exograd SAS.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os

from .account import *
from .api_client import *
from .api_object import *
from .client import *
from .context import *
from .environment import *
from .pagination import *

ca_bundle_path = os.path.join(os.path.dirname(__file__), "data", "cacert.pem")

# https://www.exograd.com/resources/pki/
public_key_pins = [
    "820df1ed4e14ad67d352960dcbdc0bdbe198390862ddf8395139f9a7303aee07"
]



# Copyright (c) 2022 Exograd SAS.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from typing import Any, Dict

from eventline.api_object import ReadableAPIObject, SerializableAPIObject


class Project(ReadableAPIObject):
    """A project."""

    def __init__(self) -> None:
        super().__init__("project")

    def _read(self, data: Dict[str, Any]) -> None:
        self.id_ = self._read_string(data, "id")
        self.org_id = self._read_string(data, "org_id")
        self.name = self._read_string(data, "name")


class NewProject(SerializableAPIObject):
    """A new project."""

    def __init__(self, name: str) -> None:
        super().__init__("new_project")
        self.name = name

    def _serialize(self) -> Dict[str, Any]:
        data = {"name": self.name}
        return data


class ProjectUpdate(SerializableAPIObject):
    """An update applied to a project."""

    def __init__(self, name: str) -> None:
        super().__init__("project_update")
        self.name = name

    def _serialize(self) -> Dict[str, Any]:
        data = {"name": self.name}
        return data
